#include "ArgStr.h"
#include <String.h>
#include <math.h>
#include <stddef.h>
#include <memory.h>
#include "RexxNumber.h"
#include "Implementation.h"
#include <assert.h>

ArgString::ArgString() : length(0), state(0), number(0) {
    // invalid state
    string = new char[1];
    string[0] = '\0';
}

// ncmp taken from libg++ String class
inline static int ncmp(const char* a, int al, const char* b, int bl) {
    int n = (al <= bl)? al : bl;
    signed char diff; 
    while (n-- > 0) if ((diff = *a++ - *b++) != 0) return diff;
    return al - bl;
}

int ArgString::operator==(ArgString& a) const {
    return (ncmp(string, length, a.string, a.length) == 0);
}

int ArgString::operator!=(ArgString& a) const {
    return (ncmp(string, length, a.string, a.length) != 0);
}

int ArgString::operator>(ArgString& a) const {
    return (ncmp(string, length, a.string, a.length) > 0);
}

int ArgString::operator<(ArgString& a) const {
    return (ncmp(string, length, a.string, a.length) < 0);
}

int ArgString::operator>=(ArgString& a) const {
    return (ncmp(string, length, a.string, a.length) >= 0);
}

int ArgString::operator<=(ArgString& a) const {
    return (ncmp(string, length, a.string, a.length) <= 0);
}

ArgString::ArgString(const char* s, int l) : number(0) {
    string = new char[l+1];
    memcpy(string, s, l);
    string[l] = '\0';
    length = l;
    state = 1; // valid state
}

ArgString::ArgString(const char c) : number(0) {
    length = 1;
    string = new char[2];
    string[0] = c;
    string[1] = '\0';
    state = 1;
}

ArgString::ArgString(const char* s) : number(0) {
    string = new char[strlen(s)+1];
    strcpy(string, s);
    length = strlen(s);
    state = 1; // valid state
}

ArgString::ArgString(double d) : number(0) {
    static char buff[128];
    sprintf(buff, "%lg", d);
    string = new char[strlen(buff)+1];
    strcpy(string, buff);
    length = strlen(buff);
    state = 1; // valid state
}

ArgString::ArgString(int i) : number(0) {
    double d = (double)i;
    static char buff[128];
    sprintf(buff, "%lg", d);
    string = new char[strlen(buff)+1];
    strcpy(string, buff);
    length = strlen(buff);
    state = 1; // valid state
}

ArgString::~ArgString() {
    delete [length+1] string;
    delete number;
}

// copy constructor
ArgString::ArgString(ArgString& a) {
    string = new char[a.length+1];
    memcpy(string, a.string, a.length);
    string[a.length] = '\0';
    length = a.length;
    state = a.state;
    if (a.number) {
	number = new RexxNumber(*a.number); // make a copy of it
    } else {
	number = 0;
    }
}

ArgString& ArgString::operator=(ArgString& a) {
    if (&a == this) return *this;

    if (length != a.length) {
	// adjust sizes
	delete [length+1] string;

	string = new char[a.length+1];
    }

    memcpy(string, a.string, a.length);
    string[a.length] = '\0';
    length = a.length;
    state = a.state;
    INVALIDATE_NUMBER;
    if (a.number) {
	number = new RexxNumber(*a.number); // make a copy of it
    } else {
	number = 0;
    }
	
    return *this;
}

ArgString& ArgString::operator+=(ArgString& a) {
    char* oldstring = string; // save the old string
    int oldlength = length; // save the old length

    string = new char[length + a.length + 1];
    memcpy(string, oldstring, length);
    memcpy(&string[length], a.string, a.length);
    length = length + a.length;
    string[length] = '\0';
    state = a.state || state; // if either is valid, we become valid

    // we needed to delay this because 'a' might be 'this'
    delete [oldlength] oldstring;
    INVALIDATE_NUMBER;

    return *this;
}

ArgString& ArgString::operator=(char* c) {
    delete [length+1] string;
    INVALIDATE_NUMBER;
    length = strlen(c);
    string = new char[length + 1];
    memcpy(string, c, length);
    string[length] = '\0';
    state = 1;
    return *this;
}

ArgString& ArgString::operator=(int i) {
    delete [length+1] string;
    INVALIDATE_NUMBER;
    static char buff[32];
    sprintf(buff, "%d", i);

    length = strlen(buff);
    string = new char[length + 1];

    memcpy(string, buff, length);
    string[length] = '\0';
    state = 1;
    return *this;
}

void ArgString::Invalidate() {
    INVALIDATE_NUMBER;
    delete string;
    string = new char[1];
    string[0] = '\0';
    length = 0;
    state = 0;
}


