#ifndef _ARGSTRING_H
#pragma once
#define _ARGSTRING_H

#include "RexxNumber.h"
#include "Implementation.h"
#include <assert.h>

class ArgString {
  private:
    int length;
    // string is always null-terminated -- length doesn't include \0
    char* string;
    int state; // whether we are uninitialized or what
    RexxNumber* number; // this will only be valid after a successful call to IsANumber()
  public:
    ArgString(); // create a new uninitialized ArgString
    ArgString(const char*, int);
    ArgString(const char*);
    ArgString(const char);
    ArgString(double);
    ArgString(int);
    ~ArgString();

    ArgString(ArgString&); // copy constructor

    int operator==(ArgString&) const;
    int operator>(ArgString&) const;
    int operator<(ArgString&) const;
    int operator>=(ArgString&) const;
    int operator<=(ArgString&) const;
    int operator!=(ArgString&) const;

    ArgString& operator+=(ArgString&);
    
    ArgString& operator=(ArgString&); // assignment operator
    ArgString& operator=(char*);
    ArgString& operator=(int);
    
    inline int Length() const;
    // isa type functions
    inline int IsANumber() const;
    inline int IsAWhole() const;
    inline int IsABoolean() const; // check if 0 or 1

    inline int IsValid() const; // check if this is a valid argstring
    void Invalidate(); // invalidate it
    // access functions
    inline const char* String() const;
    inline int Boolean() const;
    inline long Integer() const;
    inline const RexxNumber& Number() const;

    inline operator const char*() const; // same as String()

    inline char& operator[](int);
    inline const char& operator[](int) const;

    friend const ArgString& Form(const RexxNumber&, int, int);
};

#define INVALIDATE_NUMBER {delete number; number = 0;}

inline int ArgString::IsANumber() const {
    extern int validnumber(const ArgString&);
    if (number) return 1; // we already know this is valid.  In fact, we have already converted it's value into a usable form!
    if (!IsValid()) return 0;

    if (validnumber(*this)) {
	ArgString* lthis = (ArgString*)this;
	lthis->number = new RexxNumber(*this, MaxDigits + 1); // number is now valid!
	return 1;
    }
    return 0;
}

inline int ArgString::IsAWhole() const {
    if (!IsValid()) return 0;
    if (IsANumber()) return number->IsAWhole();
    return 0;
}    

inline int ArgString::IsABoolean() const {
    if (!IsValid()) return 0;
    if (IsAWhole()) {
	int i = Integer();
	return (i == 0 || i == 1);
    }
    return 0;
}

inline int ArgString::Length() const {
    return length;
}
    
inline const char* ArgString::String() const {
    return string;
}

inline const RexxNumber& ArgString::Number() const {
    assert(IsANumber());
    return *number;
}

inline long ArgString::Integer() const {
    assert(IsANumber());
    return number->Whole();
}

inline int ArgString::Boolean() const {
    assert(IsABoolean());
    return (int)(number->Whole() == 1);
}

inline int ArgString::IsValid() const {
    return state;
}

inline ArgString::operator const char*() const {
    return string;
}

inline char& ArgString::operator[](int index) {
    assert(index >= 0 && index <= length);
    INVALIDATE_NUMBER;
    return string[index];
}

inline const char& ArgString::operator[](int index) const {
    assert(index >= 0 && index <= length);
    return string[index];
}

#endif


