#include "ArgStr.h"
#include "Trace.h"
#include "Clause.h"
#include "Express.h"
#include "rexxparse.tok.h"
#include <stream.h>
#include "SymTabs.h"
#include "RexxClauses.h"
#include "support.h"
#include "MemoryFile.h"
#include "Glob.h"
#include "RexxStream.h"

extern Glob glob;

extern int LoadTokens();
extern int LoadClauses(int&);
Clause* GetClause();

//

// Clause automatically pick up glob.LineNumber when they are constructed
Clause::Clause() : linenumber(0), clausenumber(0) {
    SetID(0);
}

Clause::~Clause() {}

int Clause::Execute() {
    SetError(EInterpretationError);
    return 0;
}

int Clause::NoExecute() {
    SetError(EInterpretationError);
    return 0;
}

void Clause::SetID(int id) {
    ID = id;
}

void Clause::SetLineNumber(int l) {
    linenumber = l;
}

void Clause::SetClauseNumber(int c) {
    clausenumber = c;
}

void Clause::TraceBefore() {
    ArgString temp((int)linenumber);
    temp += " *-* ";
    fprintf(stderr, "%s\n", (char*)temp);
}

//

NopClause::NopClause() {
    SetID(TOK_NOP);
}

NopClause::~NopClause() {}

int NopClause::Execute() {
    return 1;
}

int NopClause::NoExecute() {
    return 1;
}

//int NopClause::IsA(int type) {
//    return type == TOK_NOP;
//}

//int NopClause::GetClassId() {
//    return TOK_NOP;
//}

//

OptionsClause::OptionsClause() : exp(0) {
    SetID(TOK_OPTIONS);
}

OptionsClause::~OptionsClause() {
    delete exp;
}

int OptionsClause::Execute() {
    return 1;
}

int OptionsClause::NoExecute() {
    return 1;
}

//int OptionsClause::IsA(int type) {
//    return type == TOK_OPTIONS;
//}

//int OptionsClause::GetClassId() {
//    return TOK_OPTIONS;
//}

void OptionsClause::SetExpression(Expression* e) {
    exp = e;
}

//

PushClause::PushClause() : exp(0) {
    SetID(TOK_PUSH);
}

PushClause::~PushClause() {
    delete exp;
}

int PushClause::Execute() {
    if (exp) {
	ArgString& result = exp->Evaluate();
	if (!result.IsValid()) return 0;
	glob.GlobalQueue.Push(result);
    } else {
	glob.GlobalQueue.Push(ArgString(""));
    }
    return 1;
}

int PushClause::NoExecute() {
    return 1;
}

//int PushClause::IsA(int type) {
//    return type == TOK_PUSH;
//}

//int PushClause::GetClassId() {
//    return TOK_PUSH;
//}

void PushClause::SetExpression(Expression* e) {
    exp = e;
}

//

QueueClause::QueueClause() : exp(0) {
    SetID(TOK_QUEUE);
}

QueueClause::~QueueClause() {}

int QueueClause::Execute() {
    if (exp) {
	ArgString& result = exp->Evaluate();
	if (!result.IsValid()) return 0;
	glob.GlobalQueue.Enq(result);
    } else {
	glob.GlobalQueue.Enq(ArgString(""));
    }
    return 1;
}

int QueueClause::NoExecute() {
    return 1;
}

//int QueueClause::IsA(int type) {
//    return type == TOK_QUEUE;
//}

//int QueueClause::GetClassId() {
//    return TOK_QUEUE;
//}

void QueueClause::SetExpression(Expression* e) {
    exp = e;
}

//

NumericClause::NumericClause() : formexp(0), digitsexp(0), fuzzexp(0) {
    SetID(TOK_NUMERIC);
}

NumericClause::~NumericClause() {}

int NumericClause::Execute() {
    if (fuzzexp) {
	ArgString& result = fuzzexp->Evaluate();
	if (!result.IsValid()) return 0;
	if (!result.IsAWhole()) {
	    SetError(EInvalidWholeNumber);
	    return 0;
	}
	int fuzz = result.Integer();
	if (fuzz < 0 || fuzz > GetDigits() - 1) {
	    SetError(EInvalidWholeNumber);
	    return 0;
	}
	::SetFuzz(fuzz);
    } else if (digitsexp) {
	ArgString& result = digitsexp->Evaluate();
	if (!result.IsValid()) return 0;
	if (!result.IsAWhole()) {
	    SetError(EInvalidWholeNumber);
	    return 0;
	}
	int digits = result.Integer();
	if (digits < 0 || digits <= GetFuzz()) {
	    SetError(EInvalidWholeNumber);
	    return 0;
	}
	::SetDigits(digits);
    } else if (formexp) {
	ArgString& result = formexp->Evaluate();
	if (!result.IsValid()) return 0;
	if (result == "SCIENTIFIC") ::SetForm(0);
	else if (result == "ENGINEERING") ::SetForm(1);
	else {
	    SetError(EInvalidNumericForm);
	    return 0;
	}
    } else {
	SetError(EInterpretationError);
	return 0;
    }
    return 1;
}

int NumericClause::NoExecute() {
    return 1;
}

//int NumericClause::IsA(int type) {
//    return type == TOK_NUMERIC;
//}

//int NumericClause::GetClassId() {
//    return TOK_NUMERIC;
//}

void NumericClause::SetForm(Expression* e) {
    formexp = e;
}

void NumericClause::SetDigits(Expression* e) {
    digitsexp = e;
}

void NumericClause::SetFuzz(Expression* e) {
    fuzzexp = e;
}

//

TraceClause::TraceClause() : exp(0) {
    SetID(TOK_TRACE);
}

TraceClause::~TraceClause() {}

int TraceClause::Execute() {
    return 1;
}

int TraceClause::NoExecute() {
    return 1;
}

//int TraceClause::IsA(int type) {
//    return type == TOK_TRACE;
//}

//int TraceClause::GetClassId() {
//    return TOK_TRACE;
//}

void TraceClause::SetExpression(Expression* e) {
    exp = e;
}

//

AddressClause::AddressClause() : command(0), exp(0) {
    SetID(TOK_ADDRESS);
}

AddressClause::~AddressClause() {}

int AddressClause::Execute() {
    if (!exp) {
	// flip addresses
	ArgString temp = GetCurrentAddress();
	SetCurrentAddress(GetPreviousAddress());
	SetPreviousAddress(temp);
	return 1;
    }

    if (command) {
	// save the currentaddress
	ArgString save = GetCurrentAddress();
	// set the currentaddress to environment
	SetCurrentAddress(environment);
	CommandClause* localclause = new CommandClause;
	localclause->SetExpression(exp);
	int ret = localclause->Execute(); // execute the command
	delete localclause;
	SetCurrentAddress(save);
	return ret;
    } else {
	// set the currentaddress
	ArgString& result = exp->Evaluate();
	if (!result.IsValid()) return 0;
	SetPreviousAddress(GetCurrentAddress());
	SetCurrentAddress(result);
    }
    return 1;
}

int AddressClause::NoExecute() {
    return 1;
}

//int AddressClause::IsA(int type) {
//    return type == TOK_ADDRESS;
//}

//int AddressClause::GetClassId() {
//    return TOK_ADDRESS;
//}

void AddressClause::SetCommand() {
    command = 1;
}

void AddressClause::SetEnvironment(ArgString a) {
    environment = a;
}

void AddressClause::SetExpression(Expression* e) {
    exp = e;
}

//

BreakClause::BreakClause() {
    SetID(TOK_BREAK);
}

BreakClause::~BreakClause() {}

int BreakClause::Execute() {
    SetBreak();
    return 0;
}

int BreakClause::NoExecute() {
    return 1;
}

//int BreakClause::IsA(int type) {
//    return type == TOK_BREAK;
//}

//int BreakClause::GetClassId() {
//    return TOK_BREAK;
//}

//

DropClause::DropClause() : count(0) {
    SetID(TOK_DROP);
}

DropClause::~DropClause() {}

int DropClause::Execute() {
    String drop;
    ArgString referencea;
    for (int i = 0; i < count; i++) {
	if (dropisref[i]) {
	    referencea = glob.SymbolTables.ReadAny(String((char*)droplist[i]));
	    if (glob.UninitializedAccess && SignalOn(SigNoValue)) {
		SetError(ENoValue);
		SetSignalDescription(SigNoValue, referencea);
		return 0;
	    }
	    drop = drop + referencea;
	    drop += " "; // add trailing space
	} else {
	    drop = drop + droplist[i].String();
	    drop += " "; // add a trailing space
	    }
    }
    if (!glob.SymbolTables.Drop(drop)) { // drop the list
	return 0;
    }
    return 1;
}

int DropClause::NoExecute() {
    return 1;
}

//int DropClause::IsA(int type) {
//    return type == TOK_DROP;
//}

//int DropClause::GetClassId() {
//    return TOK_DROP;
//}

void DropClause::CheckBound(int i) {
    if (i > MaxDrops) {
	cout << "DropClause::CheckBound() > MaxDrops!\n";
	abort();
    }
}

void DropClause::AddVariable(ArgString a) {
    CheckBound(count);
    droplist[count] = a;
    dropisref[count] = 0;
    count++;
}

void DropClause::AddVariableRef(ArgString a) {
    CheckBound(count);
    droplist[count] = a;
    dropisref[count] = 1;
    count++;
}

//

LeaveClause::LeaveClause() {
    SetID(TOK_LEAVE);
}

LeaveClause::~LeaveClause() {}

int LeaveClause::Execute() {
    SetLeave(variable);
    return 0;
}

int LeaveClause::NoExecute() {
    return 1;
}

//int LeaveClause::IsA(int type) {
//    return type == TOK_LEAVE;
//}

//int LeaveClause::GetClassId() {
//    return TOK_LEAVE;
//}

void LeaveClause::SetVariable(ArgString a) {
    variable = a;
}

ArgString LeaveClause::GetVariable() {
    return variable;
}

//

IterateClause::IterateClause() {
    SetID(TOK_ITERATE);
}

IterateClause::~IterateClause() {}

int IterateClause::Execute() {
    SetIterate(variable);
    return 0;
}

int IterateClause::NoExecute() {
    return 1;
}

//int IterateClause::IsA(int type) {
//    return type == TOK_ITERATE;
//}

//int IterateClause::GetClassId() {
//    return TOK_ITERATE;
//}

void IterateClause::SetVariable(ArgString a) {
    variable = a;
}

ArgString IterateClause::GetVariable() {
    return variable;
}

//

ParseClause::ParseClause() : currenttemplate(0), currentposition(0), upper(0) {
    SetID(TOK_PARSE);
    for (int i = 0; i < MaxTemplates; i++)
	for (int j = 0; j < MaxPatterns; j++)
	    ttype[i][j] = PDone;
}


ParseClause::~ParseClause() {}

int ParseClause::Execute() {
    String s;
    String m;
    String n;
    String reference;
    ArgString ma;
    ArgString na;
    ArgString referencea;
    // go through each template
    ArgString value = GetFirst();
   if (!value.IsValid()) return 0;

    for (int i = 0; i < MaxTemplates; i++) {
	s = value.String();
	if (upper) s.upcase();
	int leftmatch = 0;
	int rightmatch = s.length();
	int maxright = rightmatch; // maximum we can move to the right
	// go through each token in template
	for (int j = 0; j < MaxPatterns; j++) {
	    if (ttype[i][j] == PDone) break; // we're done with this template
	    m = ptemplate[i][j].String();
	    ma = ptemplate[i][j];
	    n = ptemplate[i][j+1].String();
	    na = ptemplate[i][j+1];
	    switch (ttype[i][j]) {
	      case PVariable :
	      case PPeriod :
	        switch (ttype[i][j+1]) {
		  case PVariable :
		  case PPeriod :
		    // make sure we start on some character
		    const char* const cp = s;
		    while (cp[leftmatch] && cp[leftmatch] == ' ') leftmatch++;
		    if (leftmatch > maxright) leftmatch = maxright;
		    // scan for a space
		    rightmatch = leftmatch; // start scanning from our current position
		    while (cp[rightmatch] && cp[rightmatch] != ' ') rightmatch++;
		    if (rightmatch > maxright) rightmatch = maxright;
		    break;
		  case PString :
		    // need to match this string
		    if (n == "") rightmatch = maxright;
		    else {
			rightmatch = s.index(n, leftmatch);
			if (rightmatch == -1) rightmatch = maxright;
		    }
		    break;
		  case PAbsoluteInteger :
		    if (!na.IsAWhole()) {
			SetError(EInvalidWholeNumber);
			return 0;
		    }
		    rightmatch = na.Integer() - 1;
		    if (rightmatch > maxright) rightmatch = maxright;
		    if (rightmatch < leftmatch) rightmatch = maxright;
		    break;
		  case PPositiveInteger :
		    if (!na.IsAWhole()) {
			SetError(EInvalidWholeNumber);
			return 0;
		    }
		    rightmatch = leftmatch + na.Integer();
		    if (rightmatch > maxright) rightmatch = maxright;
		    break;
		  case PNegativeInteger :
		    if (!na.IsAWhole()) {
			SetError(EInvalidWholeNumber);
			return 0;
		    }
		    rightmatch = leftmatch - na.Integer();
		    if (rightmatch < leftmatch) rightmatch = maxright;
		    break;
		  case PAbsoluteReference :
		    referencea = glob.SymbolTables.ReadAny(n);
		    if (glob.UninitializedAccess && SignalOn(SigNoValue)) {
			SetError(ENoValue);
			SetSignalDescription(SigNoValue, referencea);
			return 0;
		    }
		    if (!referencea.IsAWhole()) {
			SetError(EInvalidWholeNumber);
			return 0;
		    }
		    rightmatch = referencea.Integer() - 1;
		    if (rightmatch > maxright) rightmatch = maxright;
		    if (rightmatch < leftmatch) rightmatch = maxright;
		    break;
		  case PPositiveReference :
		    referencea = glob.SymbolTables.ReadAny(n);
		    if (glob.UninitializedAccess && SignalOn(SigNoValue)) {
			SetError(ENoValue);
			SetSignalDescription(SigNoValue, referencea);
			return 0;
		    }
		    if (!referencea.IsAWhole()) {
			SetError(EInvalidWholeNumber);
			return 0;
		    }
		    rightmatch = leftmatch + referencea.Integer();
		    if (rightmatch > maxright) rightmatch = maxright;
		    break;
		  case PNegativeReference :
		    referencea = glob.SymbolTables.ReadAny(n);
		    if (glob.UninitializedAccess && SignalOn(SigNoValue)) {
			SetError(ENoValue);
			SetSignalDescription(SigNoValue, referencea);
			return 0;
		    }
		    if (!referencea.IsAWhole()) {
			SetError(EInvalidWholeNumber);
			return 0;
		    }
		    rightmatch = leftmatch - referencea.Integer();
		    if (rightmatch < leftmatch) rightmatch = maxright;
		    break;
		  case PStringReference :
		    referencea = glob.SymbolTables.ReadAny(n);
		    if (glob.UninitializedAccess && SignalOn(SigNoValue)) {
			SetError(ENoValue);
			SetSignalDescription(SigNoValue, referencea);
			return 0;
		    }
		    reference = referencea.String();
		    if (reference == "") rightmatch = maxright;
		    else {
			rightmatch = s.index(reference, leftmatch);
			if (rightmatch == -1) rightmatch = maxright;
		    }
		    break;
		  case PDone :
		    rightmatch = maxright;
		    break;
		};
	        {
		String temp;
		if ((rightmatch - leftmatch) == 0) temp = "";
		else temp = s.at(leftmatch, (rightmatch - leftmatch));
		leftmatch = rightmatch;
		// only write if we are not a 'period'
		if (ttype[i][j] != PPeriod) 
		    glob.SymbolTables.WriteAny(m, ArgString((char*)temp));
	        }
		break;
	      case PString :
		// need to match this string
		if (m == "") leftmatch = maxright;
		else {
		    leftmatch = s.index(m, leftmatch);
		    if (leftmatch == -1) leftmatch = maxright;
		    else leftmatch += m.length();
		}
		break;
	      case PAbsoluteInteger :
		if (!ma.IsAWhole()) {
		    SetError(EInvalidWholeNumber);
		    return 0;
		}
		leftmatch = ma.Integer() - 1;
		if (leftmatch > maxright) leftmatch = maxright;
		if (leftmatch < 0) leftmatch = 0;
		break;
	      case PPositiveInteger :
		if (!ma.IsAWhole()) {
		    SetError(EInvalidWholeNumber);
		    return 0;
		}
		leftmatch += ma.Integer();
		if (leftmatch > maxright) leftmatch = maxright;
		break;
	      case PNegativeInteger :
		if (!ma.IsAWhole()) {
		    SetError(EInvalidWholeNumber);
		    return 0;
		}
		leftmatch -= ma.Integer();
		if (leftmatch < 0) leftmatch = 0;
		break;
	      case PAbsoluteReference :
		referencea = glob.SymbolTables.ReadAny(m);
		if (glob.UninitializedAccess && SignalOn(SigNoValue)) {
		    SetError(ENoValue);
		    SetSignalDescription(SigNoValue, referencea);
		    return 0;
		}
		if (!referencea.IsAWhole()) {
		    SetError(EInvalidWholeNumber);
		    return 0;
		}
		leftmatch = referencea.Integer() - 1;
		if (leftmatch > maxright) leftmatch = maxright;
		if (leftmatch < 0) leftmatch = 0;
		break;
	      case PPositiveReference :
		// need to access something
		referencea = glob.SymbolTables.ReadAny(m);
		if (glob.UninitializedAccess && SignalOn(SigNoValue)) {
		    SetError(ENoValue);
		    SetSignalDescription(SigNoValue, referencea);
		    return 0;
		}
		if (!referencea.IsAWhole()) {
		    SetError(EInvalidWholeNumber);
		    return 0;
		}
		leftmatch += referencea.Integer();
		if (leftmatch > maxright) leftmatch = maxright;
		break;
	      case PNegativeReference :
		// need to access something
		referencea = glob.SymbolTables.ReadAny(m);
		if (glob.UninitializedAccess && SignalOn(SigNoValue)) {
		    SetError(ENoValue);
		    SetSignalDescription(SigNoValue, referencea);
		    return 0;
		}
		if (!referencea.IsAWhole()) {
		    SetError(EInvalidWholeNumber);
		    return 0;
		}
		leftmatch -= referencea.Integer();
		if (leftmatch < 0) leftmatch = 0;
		break;
	      case PStringReference :
		// need to access something
		referencea = glob.SymbolTables.ReadAny(m);
		if (glob.UninitializedAccess && SignalOn(SigNoValue)) {
		    SetError(ENoValue);
		    SetSignalDescription(SigNoValue, referencea);
		    return 0;
		}
		reference = referencea.String();
		if (reference == "") leftmatch = maxright;
		else {
		    leftmatch = s.index(reference, leftmatch);
		    if (leftmatch == -1) leftmatch = maxright;
		    else leftmatch += reference.length();
		}
		break;
	    };
	}
        if (i < (MaxTemplates - 1)) {
	    value = GetNext();
	    if (!value.IsValid()) return 0;
	}
    }
    return 1;
}

int ParseClause::NoExecute() {
    return 1;
}

//int ParseClause::IsA(int type) {
//    return type == TOK_PARSE;
//}

//int ParseClause::GetClassId() {
//    return TOK_PARSE;
//}

void ParseClause::CheckBound(int i, int j) {
    if (i > MaxTemplates) {
	cout << "ParseClause::CheckBound() > MaxTemplates!\n";
	abort();
    }
    if (i > MaxPatterns) {
	cout << "ParseClause::CheckBound() > MaxPatterns!\n";
	abort();
    }
}

void ParseClause::AddPattern(ArgString a, ParseType t) {
    CheckBound(currenttemplate, currentposition);
    ptemplate[currenttemplate][currentposition] = a;
    ttype[currenttemplate][currentposition] = t;
    currentposition++;
}

void ParseClause::NextTemplate() {
    currenttemplate++;
    currentposition = 0;
    CheckBound(currenttemplate, currentposition);
}

void ParseClause::Upper() {
    upper = 1;
}

ArgString ParseClause::GetFirst() {
    SetError(EInterpretationError);
    return ArgString();
}

ArgString ParseClause::GetNext() {
    SetError(EInterpretationError);
    return ArgString();
}

//

ParseVarClause::ParseVarClause() {
}

ParseVarClause::~ParseVarClause() {
}

void ParseVarClause::SetVariable(ArgString a, int s) {
    variable = a;
    secondary = s;
}

ArgString ParseVarClause::GetFirst() {
    String s(variable.String());
    value = glob.SymbolTables.ReadAny(s);
    if (glob.UninitializedAccess) {
	if (SignalOn(SigNoValue)) {
	    // it is only an signal if someone did a SIGNAL ON NOVALUE
	    SetError(ENoValue);
	    SetSignalDescription(SigNoValue, value);
	    return ArgString();
	}
    }
    return value;
}

ArgString ParseVarClause::GetNext() {
    return ArgString("");
}

//

ParseVersionClause::ParseVersionClause() {
}

ParseVersionClause::~ParseVersionClause() {
}

ArgString ParseVersionClause::GetFirst() {
    return GetVersion();
}

ArgString ParseVersionClause::GetNext() {
    return ArgString("");
}

//

ParseSourceClause::ParseSourceClause() {
}

ParseSourceClause::~ParseSourceClause() {
}

ArgString ParseSourceClause::GetFirst() {
    return GetSource();
}

ArgString ParseSourceClause::GetNext() {
    return ArgString("");
}

//

ParseLineinClause::ParseLineinClause() {
}

ParseLineinClause::~ParseLineinClause() {
}

ArgString ParseLineinClause::GetFirst() {
    return GetLinein();
}

ArgString ParseLineinClause::GetNext() {
    return ArgString("");
}

//

ParseExternalClause::ParseExternalClause() {
}

ParseExternalClause::~ParseExternalClause() {
}

ArgString ParseExternalClause::GetFirst() {
    return GetExternal();
}

ArgString ParseExternalClause::GetNext() {
    return ArgString("");
}

//

ParsePullClause::ParsePullClause() {
}

ParsePullClause::~ParsePullClause() {
}

ArgString ParsePullClause::GetFirst() {
    return GetPull();
}

ArgString ParsePullClause::GetNext() {
    return ArgString("");
}

//

ParseValueClause::ParseValueClause() : exp(0) {
}

ParseValueClause::~ParseValueClause() {
}

ArgString ParseValueClause::GetFirst() {
    if (!exp) {
	SetError(EInterpretationError);
	return ArgString();
    }
    ArgString& result = exp->Evaluate();
    return result;
}

ArgString ParseValueClause::GetNext() {
    return ArgString("");
}

void ParseValueClause::SetExpression(Expression* e) {
    exp = e;
}

//

ParseArgClause::ParseArgClause() {
}

ParseArgClause::~ParseArgClause() {
}

ArgString ParseArgClause::GetFirst() {
    number = 0;
    ArgString result;
    if (ArgsOmitted(number)) {
	result = ArgString("");
    } else {
	result = Args(number);
    }
    number++;
    return result;
}

ArgString ParseArgClause::GetNext() {
    ArgString result;
    if (ArgsOmitted(number)) {
	result = ArgString("");
    } else {
	result = Args(number);
    }
    number++;
    return result;
}

//

SelectClause::SelectClause() {
    SetID(TOK_SELECT);
}

//int SelectClause::IsA(int type) {
//    return type == TOK_SELECT;
//}

//int SelectClause::GetClassId() {
//    return TOK_SELECT;
//}

int SelectClause::Execute() {
    Clause* localclause;
    int shouldexecute = 1;
    int atleastonewhen = 0;
    int ret;
    // we need to first process all the whens
    do {
	localclause = GetClause();
	if (localclause->IsA(TOK_WHEN)) {
	    atleastonewhen = 1;
	    if (shouldexecute) {
		ret = ((WhenClause*)localclause)->PrivateExecute();
		if (!ret) {
		    if (IsEof()) SetError(EIncompleteDoSelectIf);
		    return 0;
		}
		if (ret == 2) shouldexecute = 0; // this means that
		   // we already took this branch
	    } else {
		ret = ((WhenClause*)localclause)->PrivateNoExecute();
		if (!ret) {
		    if (IsEof()) SetError(EIncompleteDoSelectIf);
		    return 0;
		}
	    }
	} else {
	    break;
	}
    } while (1);
    if (!atleastonewhen) {SetError(EWhenOrOtherwiseExpected); return 0;}
    // now we either have a TOK_OTHERWISE or a TOK_END

    if (localclause->IsA(TOK_END)) {
	if (shouldexecute) {
	    SetError(EWhenOrOtherwiseExpected);
	    return 0;
	} else {
	    return 1;
	}
    }

    if (localclause->IsA(TOK_OTHERWISE) && shouldexecute) {
	do {
	    localclause = GetClause();
	    if (localclause->IsA(TOK_END)) return 1;
	    ret = localclause->Execute();
	    if (!ret) {
		if (IsEof()) SetError(EIncompleteDoSelectIf);
		return 0;
	    }
	} while (1);
    } else if (localclause->IsA(TOK_OTHERWISE) && !shouldexecute) {
	do {
	    localclause = GetClause();
	    if (localclause->IsA(TOK_END)) return 1;
	    ret = localclause->NoExecute();
	    if (!ret) {
		if (IsEof()) SetError(EIncompleteDoSelectIf);
		return 0;
	    }
	} while (1);
    }

    // if we got here, we have an illegal instruction in our select
    SetError(EWhenOrOtherwiseExpected);
    return 0;
}

int SelectClause::NoExecute() {
    return 1;
}

//

WhenClause::WhenClause() {
    SetID(TOK_WHEN);
}

//int WhenClause::IsA(int type) {
//    return type == TOK_WHEN;
//}

//int WhenClause::GetClassId() {
//    return TOK_WHEN;
//}

int WhenClause::PrivateExecute() {
    int ret;
    if (!exp) {
	SetError(EInterpretationError);
	return 0;
    }
    ArgString& temp = exp->Evaluate();
    if (!temp.IsValid()) return 0;
    if (!temp.IsABoolean()) {
	SetError(ELogicalValueNotZeroOrOne);
	return 0;
    }
    
    // expect a then clause
    Clause* localclause = GetClause();
    if (!localclause->IsA(TOK_THEN)) {
	SetError(EThenExpected);
	return 0;
    }
    
    // execute the next clause if the expression is true
    localclause = GetClause();
    if (temp.Boolean())
	ret = localclause->Execute();
    else
	ret = localclause->NoExecute();

    if (!ret) return 0; // go back up if we failed somehow

    if (temp.Boolean()) return 2;
    else return 1;
}

int WhenClause::PrivateNoExecute() {
    int ret;
    if (!exp) {
	SetError(EInterpretationError);
	return 0;
    }
    /*
    ArgString temp = exp->Evaluate();
    if (!temp.IsValid()) return 0;
    if (!temp.IsABoolean()) return 0;
    */
    // expect a then clause
    Clause* localclause = GetClause();
    if (!localclause->IsA(TOK_THEN)) {
	SetError(EThenExpected);
	return 0;
    }

    // execute the next clause if the expression is true
    localclause = GetClause();

    ret = localclause->NoExecute();

    if (!ret) return 0; // go back up if we failed somehow
    return 1;
}

int WhenClause::Execute() {
    SetError(EUnexpectedWhenOrOtherwise);
    return 0;
}

int WhenClause::NoExecute() {
    SetError(EUnexpectedWhenOrOtherwise);
    return 0;
}

void WhenClause::SetExpression(Expression* e) {
    exp = e;
}

//

OtherwiseClause::OtherwiseClause() {
    SetID(TOK_OTHERWISE);
}

//int OtherwiseClause::IsA(int type) {
//    return type == TOK_OTHERWISE;
//}

//int OtherwiseClause::GetClassId() {
//    return TOK_OTHERWISE;
//}

int OtherwiseClause::Execute() {
    cout << "otherwise not in context of select\n";
    return 0;
}

int OtherwiseClause::NoExecute() {
    cout << "otherwise not in context of select\n";
    return 0;
}



//

CallClause::CallClause() : exp(0), TurnOn_(SigSignal), TurnOff_(SigSignal) {
    SetID(TOK_CALL);
}

CallClause::~CallClause() {
}

int CallClause::Execute() {
    if (TurnOff_) SetSignalOff(TurnOff_);
    else if (TurnOn_) SetSignalOn(TurnOn_, SignalLabel, ACall);
    else {
	if (!exp) {
	    SetError(EInterpretationError);
	    return 0;
	}
	exp->SetCall();
	ArgString& a = exp->Evaluate();
	// need to check why it returns a bad value -- can't assume that it failed
	if (!a.IsValid()) {
	    if (IsReturn()) {
		// this is OK
	    } else {
		return 0;
	    }
	    glob.SymbolTables.DropSimple("RESULT");
	} else {
	    glob.SymbolTables.Write("RESULT", a);
	}
    }
    return 1;
}

int CallClause::NoExecute() {
    return 1;
}

//int CallClause::IsA(int type) {
//    return type == TOK_CALL;
//}

//int CallClause::GetClassId() {
//    return TOK_CALL;
//}

void CallClause::TurnOff(Signals s) {
    TurnOff_ = s;
}

void CallClause::TurnOn(Signals s, ArgString l) {
    SignalLabel = l;
    TurnOn_ = s;
}

void CallClause::TurnOn(Signals s) {
    TurnOn_ = s;
}

void CallClause::SetExpression(FunctionExpression* e) {
    exp = e;
}

//

InterpretClause::InterpretClause() : exp(0) {
    SetID(TOK_INTERPRET);
}

InterpretClause::~InterpretClause() {
}

//int InterpretClause::IsA(int type) {
//    return type == TOK_INTERPRET;
//}

//int InterpretClause::GetClassId() {
//    return TOK_INTERPRET;
//}

int InterpretClause::Execute() {
    int start, previous;
    int iret;
    if (exp) {
	ArgString& ret = exp->Evaluate();
	if (ret.IsValid()) {
	    int previousline = glob.LineNumber;
	    // load up tokens
	    glob.Stage1.LoadFromString(String(ret.String()));
	    if (!LoadTokens()) return 0;
	    // parse it
	    int end;
	    start = LoadClauses(end);

	    // we don't do a SetStart() or SetEnd() here!!

	    // execute it
	    previous = glob.Clauses->GetPC();
	    glob.Clauses->SetPC(start);
	    while (1) {
		Clause* localclause = GetClause();
		iret = localclause->Execute();
		if (!iret) {
		    if (IsEof() || IsBreak()) {
			iret = 1;
		    }
		    break;
		}
	    }
		
	    // deallocate the parsed stuff
	    glob.Clauses->Free(start);
	    glob.Clauses->SetPC(previous);

	    glob.LineNumber = previousline;

	    return iret;
	} else {
	    return 0; // pass on the error code
	}
    } else {
	abort();
    }
    return 1;
}

int InterpretClause::NoExecute() {
    cout << "interpretclause::noexecute\n";
    return 1;
}

void InterpretClause::SetExpression(Expression* e) {
    exp = e;
}

//

ProcedureClause::ProcedureClause() : count(0) {
    SetID(TOK_PROCEDURE);
}

ProcedureClause::~ProcedureClause() {}

int ProcedureClause::Execute() {
    if (HasSymbolTable()) {
	SetError(EUnexpectedProcedure);
	return 0;
    } else {
	glob.SymbolTables.PushSymbolTable();
	HasSymbolTable(1);
	// see if there is any expose list
	String expose;
	ArgString referencea;
	for (int i = 0; i < count; i++) {
	    if (exposeisref[i]) {
		referencea = glob.SymbolTables.ReadAny(String((char*)exposelist[i]));
		if (glob.UninitializedAccess && SignalOn(SigNoValue)) {
		    SetError(ENoValue);
		    SetSignalDescription(SigNoValue, referencea);
		    return 0;
		}
		expose = expose + referencea;
		expose += " "; // add trailing space
	    } else {
		expose = expose + exposelist[i].String();
		expose += " "; // add a trailing space
	    }
	}
	if (!glob.SymbolTables.Expose(expose)) { // expose the list
	    return 0;
	}
    }
    return 1;
}

int ProcedureClause::NoExecute() {
    return 1;
}

/*
int ProcedureClause::IsA(int type) {
    return type == TOK_PROCEDURE;
}

int ProcedureClause::GetClassId() {
    return TOK_PROCEDURE;
}
*/

void ProcedureClause::CheckBound(int i) {
    if (i > MaxExposes) {
	cout << "ProcedureClause::CheckBound() > MaxExposes!\n";
	abort();
    }
}

void ProcedureClause::AddVariable(ArgString a) {
    CheckBound(count);
    exposelist[count] = a;
    exposeisref[count] = 0;
    count++;
}

void ProcedureClause::AddVariableRef(ArgString a) {
    CheckBound(count);
    exposelist[count] = a;
    exposeisref[count] = 1;
    count++;
}

//

ErrorClause::ErrorClause() {
    SetID(TOK_ERRORCLAUSE);
}

/*
int ErrorClause::IsA(int type) {
    return type == TOK_ERRORCLAUSE;
}

int ErrorClause::GetClassId() {
    return TOK_ERRORCLAUSE;
}
*/

// these should really print appropriate messages and return correct error codes
int ErrorClause::Execute() {
    SetError(ESyntax); // I couldn't figure out what kind of error this is
    return 0;
}

int ErrorClause::NoExecute() {
    SetError(ESyntax); // I couldn't figure out what kind of error this is
    return 0;
}

//

NullClause::NullClause() {
    SetID(TOK_NULLCLAUSE);
}

/*
int NullClause::IsA(int type) {
    return type == TOK_NULLCLAUSE;
}

int NullClause::GetClassId() {
    return TOK_NULLCLAUSE;
}
*/

//

int LabelClause::Matches(String l) {
    return label == l;
}

LabelClause::LabelClause(ArgString a) {
    label = a.String(); // convert into a String
    label.gsub(RXwhite, ""); // kill white space in label
    label.upcase();
    SetID(TOK_LABEL);
}

/*
int LabelClause::IsA(int type) {
    return type == TOK_LABEL;
}

int LabelClause::GetClassId() {
    return TOK_LABEL;
}
*/

//

SignalClause::SignalClause() : exp(0), TurnOn_(SigSignal), TurnOff_(SigSignal) {
    SetID(TOK_SIGNAL);
}

SignalClause::~SignalClause() {
}

int SignalClause::Execute() {
    if (TurnOff_) SetSignalOff(TurnOff_);
    else if (TurnOn_) SetSignalOn(TurnOn_, SignalLabel);
    else {
	ArgString& a = exp->Evaluate();
	if (!a.IsValid()) return 0;
	SetSignalOn(SigSignal, a);
	SetError(ESignal); // forcefeed a signal
	return 0;
    }
    return 1;
}

int SignalClause::NoExecute() {
    return 1;
}

/*
int SignalClause::IsA(int type) {
    return type == TOK_SIGNAL;
}

int SignalClause::GetClassId() {
    return TOK_SIGNAL;
}
*/

void SignalClause::TurnOff(Signals s) {
    TurnOff_ = s;
}

void SignalClause::TurnOn(Signals s, ArgString l) {
    SignalLabel = l;
    TurnOn_ = s;
}

void SignalClause::TurnOn(Signals s) {
    TurnOn_ = s;
}

void SignalClause::SetExpression(Expression* e) {
    exp = e;
}

//

DoClause::DoClause() : Forever(0), secondary(0), To(0), By(0), While(0), Until(0), For(0), From(0), LongTo(0), LongFrom(0), LongBy(0), vsymbol(0), EndLabelChecked(0), EndLabelMatched(0) {
    SetID(TOK_DO);
}

DoClause::~DoClause() {
    delete vsymbol;
}

/*
int DoClause::IsA(int type) {
    return type == TOK_DO;
}

int DoClause::GetClassId() {
    return TOK_DO;
}
*/

ArgString& DoClause::ReadSymbol() {
    return vsymbol->Read();
//    switch (secondary) {
//      case TOK_SIMPLE : 
//	return glob.SymbolTables.Read(ssymbol);
//	break;
//      case TOK_STEM :
//	return glob.SymbolTables.ReadStem(ssymbol);
//	break;
//      case TOK_COMPOUND :
//	return glob.SymbolTables.ReadCompound(ssymbol);
//    };
    abort();
}

inline void DoClause::SetTop() {
    DoPC = glob.Clauses->GetPC();
}

inline void DoClause::JumpTop() {
    glob.Clauses->SetPC(DoPC);
}

inline int DoClause::WhileEvaluate() {
    if (While) {
	ArgString& a = While->Evaluate();
	if (!a.IsValid()) return 0;
	if (!a.IsABoolean()) {
	    SetError(ELogicalValueNotZeroOrOne);
	    return 0;
	}
	WhileInt = a.Boolean();
    }
}	

inline int DoClause::UntilEvaluate() {
    if (Until) {
	ArgString& a = Until->Evaluate();
	if (!a.IsValid()) return 0;
	if (!a.IsABoolean()) {
	    SetError(ELogicalValueNotZeroOrOne);
	    return 0;
	}
	UntilInt = a.Boolean();
    }
}	

inline int DoClause::StartEvaluate() {
    if (From) {
	ArgString& a = From->Evaluate();
	if (!a.IsValid()) return 0;
	if (!a.IsANumber()) {
	    SetError(ELogicalValueNotZeroOrOne);
	    return 0;
	}
	FromNumber = a.Number();
//	if (FromNumber.IsAWhole()) {
//	    FromLong = FromNumber.Whole();
//	    LongFrom = 1;
//	} else LongFrom = 0;

	// update symbol table with new variable
	WriteSymbol(FromNumber);

	if (To) {
	    ArgString& a = To->Evaluate();
	    if (!a.IsValid()) return 0;
	    if (!a.IsANumber()) {
		SetError(ELogicalValueNotZeroOrOne);
		return 0;
	    }
	    ToNumber = a.Number();
//	    if (ToNumber.IsAWhole()) {
//		ToLong = ToNumber.Whole();
//		LongTo = 1;
//	    } else LongTo = 0;
	}
	if (By) {
	    ArgString& a = By->Evaluate();
	    if (!a.IsValid()) return 0;
	    if (!a.IsANumber()) {
		SetError(ELogicalValueNotZeroOrOne);
		return 0;
	    }
	    ByNumber = a.Number();
	} else {
	    ByNumber = RexxNumber(1);
	}
//	if (ByNumber.IsAWhole()) {
//	    ByLong = ByNumber.Whole();
//	    LongBy = 1;
//	} else LongBy = 0;
    }
    if (For) {
	ArgString& a = For->Evaluate();
	if (!a.IsValid()) return 0;
	if (!a.IsAWhole()) {
	    SetError(EInvalidWholeNumber);
	    return 0;
	}
	ForInt = a.Integer();
	if (ForInt < 0) {
	    SetError(EInvalidWholeNumber);
    return 0;
	}
    }
    return 1;
}

inline int DoClause::EndEvaluate() {
    if (For) ForInt--;
    return 1;
}

inline int DoClause::UpdateIndex() {
    if (From) {
	ArgString& a = ReadSymbol();
	if (!a.IsValid()) {
	    return 0; // not a number error
	}
	if (!a.IsANumber()) {
	    SetError(EBadArithmeticConversion);
	    return 0; // not a number
	}
	add(a.Number(), ByNumber, GetDigits(), FromNumber);
	WriteSymbol(FromNumber);
    }
    return 1;
}

inline void DoClause::WriteSymbol(RexxNumber& Value) {
    ArgString& r = Form(Value, GetDigits());
    vsymbol->Write(r);
//    switch (secondary) {
//      case TOK_SIMPLE : 
//      glob.SymbolTables.Write(ssymbol, r);
//	break;
//      case TOK_STEM :
//	glob.SymbolTables.WriteStem(ssymbol, r);
//	break;
//      case TOK_COMPOUND :
//	glob.SymbolTables.WriteCompound(ssymbol, r);
//    };
}

inline int DoClause::StartDone() {
    if (While && !WhileInt) return 1;
    if (For && !ForInt) return 1;
    if (From) {
	// read current index variable
	if (ByNumber.sign() >= 0) {
	    if (To) {
//		if (LongFrom && LongTo) {
//		    if (FromLong > ToLong) return 1;
//		} else {
		    if (FromNumber.greater(ToNumber)) return 1;

		    // code below does 'rexx' comparison
		    // I don't think we need to do that here
		    /*
		    RexxNumber comp;
		    RexxNumber n1 = FromNumber;
		    RexxNumber n2 = ToNumber;
		    n2.negate();
		    add(n1, n2, GetDigits(), comp);
		    // if fromnumber > tonumber
		    if (comp.sign() > 0) return 1;
		    */
//		}
	    }
	} else {
	    if (To) {
//		if (LongFrom && LongTo) {
//		    if (FromLong < ToLong) return 1;
//		} else {
		    if (ToNumber.greater(FromNumber)) return 1;
		/*
		    RexxNumber comp;
		    RexxNumber n1 = FromNumber;
		    RexxNumber n2 = ToNumber;
		    n2.negate();
		    add(n1, n2, GetDigits(), comp);
		    // if fromnumber < tonumber
		    if (comp.sign() < 0) return 1;
		    */
//		}
	    }
	}
    }
    return 0;
}

inline int DoClause::IsRepetitive() {
    return (Forever || From || To || By || For || While || Until);
}

inline int DoClause::EndDone() {
    if (Until && UntilInt) return 1;
    if (For && !ForInt) return 1;
    if (!Forever && !From && !For && !While && !Until) return 1;
    return 0;
}

inline int DoClause::DoEndStuff() {
    // check the ending conditions here
    if (!EndEvaluate()) return 0;
    if (!UntilEvaluate()) return 0;
    if (EndDone()) {
	return 1;
    } else {
	if (!UpdateIndex()) return 0;
	if (!WhileEvaluate()) return 0;
	if (StartDone()) return 1;
	if (!ProcessPendingSignals()) return 0;
	JumpTop();
	return 2; // should continue!
    }
}

int DoClause::Execute() {
    Clause* localclause;
    SetTop();

    // check the starting conditions here
    if (!StartEvaluate()) return 0;
    if (!WhileEvaluate()) return 0;
    if (StartDone()) return NoExecute();

    while (1) {
	localclause = GetClause();
	if (localclause->IsA(TOK_END)) {
	    // check if the label matches ours
	    if (!EndLabelChecked) {
		EndLabelChecked = 1;
		ArgString& endlabel = ((EndClause*)localclause)->GetVariable();
		if (endlabel.IsValid()) {
		    if (symbol.IsValid()) {
			if (String(symbol.String()) != String(endlabel.String())) {
			    SetError(EInvalidEnd);
			    EndLabelMatched = 0;
			    return 0;
			}
		    } else {
			SetError(EInvalidEnd);
			EndLabelMatched = 0;
			return 0;
		    }
		}
		EndLabelMatched = 1;
	    } else {
		if (!EndLabelMatched) {
		    SetError(EInvalidEnd);
		    return 0;
		}
	    }
	    // check the ending conditions here
	    if (!EndEvaluate()) return 0;
	    if (!UntilEvaluate()) return 0;
	    if (EndDone()) {
		return 1;
	    } else {
		if (!UpdateIndex()) return 0;
		if (!WhileEvaluate()) return 0;
		if (StartDone()) return 1;
		if (!ProcessPendingSignals()) return 0;
		JumpTop();
	    }
	} else {
	    if (!localclause->Execute()) {
		// could have returned 0 because of break, leave, or iterate
		if (IsEof()) {
		    // we're not supposed to get this here
		    // this is BAD
		    SetError(EIncompleteDoSelectIf);
		    return 0;
		} else if (IsBreak()) {
		    if (!NoExecute()) return 0;
		    return 1;
		} else if (IsIterate()) {
		    if (!NoExecute()) return 0;
		    ArgString ret = ReturnValue();
		    if (ret.IsValid()) {
			if (symbol.IsValid()) {
			    if (String(ret.String()) == String(symbol.String())) {
				// iterate
				int ret = DoEndStuff();
				if (ret != 2) return ret;
			    } else {
				return 0; // let someone else catch it
			    }
			} else {
			    return 0; // let someone else catch it
			}
		    } else {
			// no name specified
			if (IsRepetitive()) {
			    // iterate
			    int ret = DoEndStuff();
			    if (ret != 2) return ret;
			} else {
			    SetError(EInvalidLeaveOrIterate);
			    return 0;
			}
		    }
		} else if (IsLeave()) {
		    if (!NoExecute()) return 0;
		    ArgString ret = ReturnValue();
		    if (ret.IsValid()) {
			if (symbol.IsValid()) {
			    if (String(ret.String()) == String(symbol.String())) return 1;
			    else return 0; // let someone else catch it
			} else {
			    return 0; // let someone else catch it
			}
		    } else {
			// no name specified
			if (IsRepetitive()) {
			    return 1;
			} else {
			    SetError(EInvalidLeaveOrIterate);
			    return 0;
			}
		    }
		} else {
		    // we couldn't handle it here
		    return 0;
		}
	    } else { // if localclause->Execute() returned properly
		// see if there are any pending signals
		if (!ProcessPendingSignals()) return 0;
	    }
	}
    }
}

int DoClause::NoExecute() {
    Clause* localclause;
    while (1) {
	localclause = GetClause();
	if (localclause->IsA(TOK_END)) {
	    if (!EndLabelChecked) {
		EndLabelChecked = 1;
		ArgString& endlabel = ((EndClause*)localclause)->GetVariable();
		if (endlabel.IsValid()) {
		    if (symbol.IsValid()) {
			if (ssymbol != String(endlabel.String())) {
			    SetError(EInvalidEnd);
			    EndLabelMatched = 0;
			    return 0;
			}
		    } else {
			SetError(EInvalidEnd);
			EndLabelMatched = 0;
			return 0;
		    }
		}
		EndLabelMatched = 1;
		return 1;
	    } else {
		return EndLabelMatched;
	    }
	}
	if (!localclause->NoExecute()) {
	    if (IsEof()) SetError(EIncompleteDoSelectIf);
		
	    return 0;
	}
    }
}

int DoClause::SetForever() {
    Forever = 1;
}

int DoClause::SetSymbol(ArgString& a, int i) {
    switch (i) {
      case TOK_SIMPLE : 
	vsymbol = new SimpleVariable(a.String());
	break;
      case TOK_STEM :
	vsymbol = new StemVariable(a.String());
	break;
      case TOK_COMPOUND :
	vsymbol = new CompoundVariable(a.String());
    };
    symbol = a;
    ssymbol = (char*)a;
//    secondary = i;
}

int DoClause::SetFrom(Expression* e) {
    From = e;
}

int DoClause::SetTo(Expression* e) {
    To = e;
}

int DoClause::SetBy(Expression* e) {
    By = e;
}

int DoClause::SetFor(Expression* e) {
    For = e;
}

int DoClause::SetWhile(Expression* e) {
    While = e;
}

int DoClause::SetUntil(Expression* e) {
    Until = e;
}

//

EndClause::EndClause() {
    SetID(TOK_END);
}

/*
int EndClause::IsA(int type) {
    return type == TOK_END;
}

int EndClause::GetClassId() {
    return TOK_END;
}
*/

int EndClause::Execute() {
    SetError(EUnexpectedOrUnmatchedEnd);
    return 0;
}

int EndClause::NoExecute() {
    SetError(EUnexpectedOrUnmatchedEnd);
    return 0;
}

void EndClause::SetVariable(ArgString& a) {
    variable = a;
}

ArgString& EndClause::GetVariable() {
    return variable;
}

//

IfClause::IfClause() {
    SetID(TOK_IF);
}

/*
int IfClause::IsA(int type) {
    return type == TOK_IF;
}

int IfClause::GetClassId() {
    return TOK_IF;
}
*/

void IfClause::SetExpression(Expression* e) {
    exp = e;
}

int IfClause::Execute() {
    int ret;
    if (!exp) {
	SetError(EInterpretationError);
	return 0;
    }
    ArgString& temp = exp->Evaluate();
    if (!temp.IsValid()) return 0;
    if (!temp.IsABoolean()) {
	SetError(ELogicalValueNotZeroOrOne);
	return 0;
    }

    // expect a then clause
    Clause* localclause = GetClause();
    if (!localclause->IsA(TOK_THEN)) {
	SetError(EThenExpected);
	return 0;
    }

    // execute the next clause if the expression is true
    localclause = GetClause();
    if (temp.Boolean())
	ret = localclause->Execute();
    else
	ret = localclause->NoExecute();

    if (!ret) {
	if (IsEof()) SetError(EIncompleteDoSelectIf);
	return 0; // go back up if we failed somehow
    }

    // expect an else
    int CurPC = glob.Clauses->GetPC(); // save our current position
    // test parse the next instruction
    localclause = GetClause();
    // if the next instruction is not an else clause, reset pc and return
    if (!localclause->IsA(TOK_ELSE)) {
	glob.Clauses->SetPC(CurPC);
	return 1;
    }
    localclause = GetClause();
    // otherwise, get execute the next instruction
    if (!temp.Boolean())
	ret = localclause->Execute();
    else
	ret = localclause->NoExecute();
    if (!ret) {
	if (IsEof()) SetError(EIncompleteDoSelectIf);
	return 0;
    }

    return 1;
}

int IfClause::NoExecute() {
    int ret;
    if (!exp) {
	SetError(EInterpretationError);
	return 0;
    }
    ArgString& temp = exp->Evaluate();
    if (!temp.IsValid()) return 0;
    if (!temp.IsABoolean()) {
	SetError(ELogicalValueNotZeroOrOne);
	return 0;
    }

    // expect a then clause
    Clause* localclause = GetClause();
    if (!localclause->IsA(TOK_THEN)) {
	SetError(EThenExpected);
	return 0;
    }

    // execute the next clause if the expression is true
    localclause = GetClause();

    ret = localclause->NoExecute();

    if (!ret) {
	if (IsEof()) SetError(EIncompleteDoSelectIf);
	return 0; // go back up if we failed somehow
    }

    // expect an else
    int CurPC = glob.Clauses->GetPC(); // save our current position
    // test parse the next instruction
    localclause = GetClause();
    // if the next instruction is not an else clause, reset pc and return
    if (!localclause->IsA(TOK_ELSE)) {
	glob.Clauses->SetPC(CurPC);
	return 1;
    }
    localclause = GetClause();

    ret = localclause->NoExecute();

    if (!ret) {
	if (IsEof()) SetError(EIncompleteDoSelectIf);
	return 0;
    }

    return 1;
}

//

ThenClause::ThenClause() {
    SetID(TOK_THEN);
}

/*
int ThenClause::IsA(int type) {
    return type == TOK_THEN;
}

int ThenClause::GetClassId() {
    return TOK_THEN;
}
*/

int ThenClause::Execute() {
    SetError(EUnexpectedThenOrElse);
    return 0;
}

int ThenClause::NoExecute() {
    SetError(EUnexpectedThenOrElse);
    return 0;
}

//

ElseClause::ElseClause() {
    SetID(TOK_ELSE);
}

/*
int ElseClause::IsA(int type) {
    return type == TOK_ELSE;
}

int ElseClause::GetClassId() {
    return TOK_ELSE;
}
*/

int ElseClause::Execute() {
    SetError(EUnexpectedThenOrElse);
    return 0;
}

int ElseClause::NoExecute() {
    SetError(EUnexpectedThenOrElse);
    return 0;
}

//

SayClause::SayClause() : exp(0) {
    SetID(TOK_SAY);
}

void SayClause::SetExpression(Expression* e) {
    exp = e;
}

/*
int SayClause::IsA(int type) {
    return type == TOK_SAY;
}

int SayClause::GetClassId() {
    return TOK_SAY;
}
*/

int SayClause::Execute() {
    if (exp) {
	ArgString& temp = exp->Evaluate();
	if (temp.IsValid()) {
	    glob.Streams["stdout"]->Lineout(temp);
//	    write(1, temp.String(), temp.Length());
//	    write(1, "\n", 1);
	} else {
	    return 0;
	}
    } else {
	glob.Streams["stdout"]->Lineout(ArgString(""));
//	write(1, "\n", 1);
    }
    return 1;
}

int SayClause::NoExecute() {
    return 1;
}

//

ExitClause::ExitClause() : exp(0) {
    SetID(TOK_EXIT);
}

/*
int ExitClause::IsA(int type) {
    return type == TOK_EXIT;
}

int ExitClause::GetClassId() {
    return TOK_EXIT;
}
*/

int ExitClause::Execute() {
    if (exp) {
	ArgString& ret = exp->Evaluate();
	if (ret.IsValid()) {
	    SetExit(ret);
	} else {
	    return 0;
	}
    } else {
	SetExit(ArgString()); // no return value
    }
//    cout << "exitclause::execute\n";
    return 0;
}

int ExitClause::NoExecute() {
    cout << "exitclause::noexecute\n";
    return 1;
}

void ExitClause::SetExpression(Expression* e) {
    exp = e;
}

//

EofClause::EofClause() {
    SetID(TOK_EOF);
}

/*
int EofClause::IsA(int type) {
    return type == TOK_EOF;
}

int EofClause::GetClassId() {
    return TOK_EOF;
}
*/

int EofClause::Execute() {
    SetEof(); // no return value
//    cout << "exitclause::execute\n";
    return 0;
}

int EofClause::NoExecute() {
//    cout << "exitclause::noexecute\n";
    SetEof();
    return 0;
}

//

ReturnClause::ReturnClause() : exp(0) {
    SetID(TOK_RETURN);
}

/*
int ReturnClause::IsA(int type) {
    return type == TOK_RETURN;
}

int ReturnClause::GetClassId() {
    return TOK_RETURN;
}
*/

int ReturnClause::Execute() {
    if (exp) {
	ArgString& ret = exp->Evaluate();
	if (ret.IsValid()) {
	    SetReturn(ret);
	} else {
	    return 0;
	}
    } else {
	SetReturn(ArgString()); // no return value
    }
//    cout << "returnclause::execute\n";
    return 0;
}

int ReturnClause::NoExecute() {
    cout << "returnclause::noexecute\n";
    return 1;
}

void ReturnClause::SetExpression(Expression* e) {
    exp = e;
}

//

//AssignmentClause::AssignmentClause(ArgString a, int s, Expression* e) :
//    symbol((char*)a), secondary(s), exp(e) {}

AssignmentClause::AssignmentClause(ArgString& a, int s, Expression* e) : exp(e), vsymbol(0) {
    switch (s) {
      case TOK_SIMPLE :
	vsymbol = new SimpleVariable(a.String());
	break;
      case TOK_STEM :
	vsymbol = new StemVariable(a.String());
	break;
      case TOK_COMPOUND :
	vsymbol = new CompoundVariable(a.String());
	break;
    };
    SetID(TOK_ASSIGNMENT);
}

AssignmentClause::~AssignmentClause() {
    delete exp;
    delete vsymbol;
}

/*
int AssignmentClause::IsA(int type) {
    return type == TOK_ASSIGNMENT;
}

int AssignmentClause::GetClassId() {
    return TOK_ASSIGNMENT;
}
*/

int AssignmentClause::Execute() {
    const ArgString& rhs = exp->Evaluate();
    if (!rhs.IsValid()) return 0;
    
    vsymbol->Write(rhs);
    
    return 1;
}

int AssignmentClause::NoExecute() {
return 1;
}

//

CommandClause::CommandClause() : exp(0) {
    SetID(TOK_COMMANDCLAUSE);
}

void CommandClause::SetExpression(Expression* e) {
    exp = e;
}

/*
int CommandClause::IsA(int type) {
    return type == TOK_COMMANDCLAUSE;
}

int CommandClause::GetClassId() {
    return TOK_COMMANDCLAUSE;
}
*/

int CommandClause::Execute() {
    extern int convertretvalue(ArgString);
    extern int searchforrexxfile(String, String, String&);
    extern ArgString externalcommand(String, ArgString);

    if (!exp) {
	SetError(EInterpretationError);
	return 0;
    }

    String currentaddress;
    currentaddress = GetCurrentAddress().String();
//    cout << "in commandclause " << currentaddress << GetCurrentAddress().String() << "\n";
    if (currentaddress != "SHELL") {
	// special trapping for non-shell addresses
	SetError(EUnknownAddress);
	return 0;
    }
    
    ArgString& temp = exp->Evaluate();
    if (!temp.IsValid()) return 0;

    //cout << "Command::" << temp.String() << "\n";
    // intercept any rexx commands before trying system()
    String command = temp.String();
    String resolved;
    if (command.index(RXwhite, 0) == 0) command = command.after(RXwhite);
    String head;
    String tail;

    if (command.contains(RXwhite)) {
	head = command.before(RXwhite);
	tail = command.after(RXwhite);
	if (tail.index(RXwhite, 0) == 0) tail = tail.after(RXwhite);
    } else {
	head = command;
	tail = "";
    }

    int ret;
    ArgString commandret;
    if (searchforrexxfile(head, glob.DefaultExtension, resolved)) {
	// we found a rexx file to execute
	if (tail == "") {
	    commandret = externalcommand(resolved, ArgString());
	} else {
	    commandret = externalcommand(resolved, ArgString((char*)tail));
	}
	if (commandret.IsValid()) {
	    ret = convertretvalue(commandret);
	} else {
	    // we should NEVER get here
	    abort();
	    if (IsReturn() || IsExit()) {
		ret = 0;
		// we got here because there was no return value supplied
	    } else {
		// print out some nasty errors
		fprintf(stderr, "+++ Error %d in line %d:%s\n",
			glob.Error, glob.LineNumber, GetErrorText(glob.Error));
		ret = 10;
	    }
	}
    } else {
	ret = system(temp.String());
	ret = ret / 0xff; // decode return code
    }

    glob.SymbolTables.Write("RC", ArgString(ret));
    if (ret == 0) return 1;
    else if (ret < 10) {
	// command error
	if (SignalOn(SigError)) {
	    if (ActionIsSignal(SigError)) {
		SetError(EError);
		SetSignalDescription(SigError, temp);
		return 0;
	    } else {
		SetSignalPending(SigError);
//		SetSignalDelayed(SigError);
		SetSignalDescription(SigError, temp);
	    }
	    // otherwise, ignore it
	}
    } else { // ret > 10
	// possibly a command failure
	if (SignalOn(SigFailure)) {
	    if (ActionIsSignal(SigFailure)) {
		SetError(EFailure);
		SetSignalDescription(SigFailure, temp);
		return 0;
	    } else {
		SetSignalPending(SigFailure);
//		SetSignalDelayed(SigFailure);
		SetSignalDescription(SigFailure, temp);
	    }
	} else if (SignalOn(SigError)) {
	    // if we aren't catching SigFailures, perhaps we're catching SigErrors
	    if (ActionIsSignal(SigError)) {
		SetError(EError);
		SetSignalDescription(SigError, temp);
		return 0;
	    } else {
		SetSignalPending(SigError);
//		SetSignalDelayed(SigError);
		SetSignalDescription(SigError, temp);
	    }
	}
	// otherwise, ignore it
    }
	
    return 1;
}

int CommandClause::NoExecute() {
    return 1;
}


