// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)
    based on code by Marc Shapiro (shapiro@sor.inria.fr)

This file is part of GNU CC.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the GNU CC General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
GNU CC, but only under the conditions described in the
GNU CC General Public License.   A copy of this license is
supposed to have been given to you along with GNU CC so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies.  
*/

#ifndef _ClausepPlex_h
#ifdef __GNUG__
#pragma once
#pragma interface
#endif
#define _ClausepPlex_h 1

#include <std.h>
#include <Pix.h>
#include "Clausep.defs.h"

// Plexes are made out of ClausepIChunks

#include <stddef.h>

class ClausepIChunk
{
//public: // kludge until C++ `protected' policies settled
protected:      

  Clausep*           data;           // data, from client

  int            base;           // lowest possible index
  int            low;            // lowest valid index
  int            fence;          // highest valid index + 1
  int            top;            // highest possible index + 1

  ClausepIChunk*     nxt;            // circular links
  ClausepIChunk*     prv;

public:

// constructors

                 ClausepIChunk(Clausep*     d,       // ptr to array of elements
                        int      base_idx, // initial indices
                        int      low_idx,  
                        int      fence_idx,
                        int      top_idx);

  virtual       ~ClausepIChunk();

// status reports

  int            size() const;    // number of slots

  virtual int    empty() const ;
  virtual int    full() const ; 

  int            can_grow_high () const ;  // there is space to add data
  int            can_grow_low () const;        
  
  int            base_index() const;   // lowest possible index;
  int            low_index() const;    // lowest actual index;
  virtual int    first_index() const;  // lowest valid index or fence if none
  virtual int    last_index() const;   // highest valid index or low-1 if none
  int            fence_index() const;  // highest actual index + 1
  int            top_index() const;    // highest possible index + 1

// indexing conversion

  int            possible_index(int i) const; // i between base and top
  int            actual_index(int i) const;   // i between low and fence
  virtual int    valid_index(int i) const;    // i not deleted (mainly for mchunks)

  int            possible_pointer(const Clausep* p) const; // same for ptr
  int            actual_pointer(const Clausep* p) const; 
  virtual int    valid_pointer(const Clausep* p) const; 

  Clausep*           pointer_to(int i) const ;   // pointer to data indexed by i
                                      // caution: i is not checked for validity
  int            index_of(const Clausep* p) const; // index of data pointed to by p
                                      // caution: p is not checked for validity

  virtual int    succ(int idx) const;     // next valid index or fence if none
  virtual int    pred(int idx) const;     // previous index or low - 1 if none

  virtual Clausep*   first_pointer() const;   // pointer to first valid pos or 0
  virtual Clausep*   last_pointer() const;    // pointer to first valid pos or 0
  virtual Clausep*   succ(Clausep*  p) const;     // next pointer or 0
  virtual Clausep*   pred(Clausep* p) const;     // previous pointer or 0

// modification

  virtual Clausep*   grow_high ();      // return spot to add an element
  virtual Clausep*   grow_low ();  

  virtual void   shrink_high ();    // logically delete top index
  virtual void   shrink_low ();     

  virtual void   clear(int lo);     // reset to empty ch with base = lo
  virtual void   cleardown(int hi); // reset to empty ch with top = hi
  void           re_index(int lo);  // re-index so lo is new low

// chunk traversal

  ClausepIChunk*     next() const;
  ClausepIChunk*     prev() const;

  void           link_to_prev(ClausepIChunk* prev);
  void           link_to_next(ClausepIChunk* next);
  void           unlink();

// state checks

  Clausep*           invalidate();     // mark self as invalid; return data
                                   // for possible deletion

  virtual int    OK() const;             // representation invariant

  volatile void   error(const char*) const;
  volatile void   empty_error() const;
  volatile void   full_error() const;
  volatile void   index_error() const;
};

// ClausepPlex is a partly `abstract' class: few of the virtuals
// are implemented at the Plex level, only in the subclasses

class ClausepPlex
{
protected:      

  ClausepIChunk*       hd;          // a chunk holding the data
  int              lo;          // lowest  index
  int              fnc;         // highest index + 1
  int              csize;       // size of the chunk

  void             invalidate();              // mark so OK() is false
  void             del_chunk(ClausepIChunk*);        // delete a chunk

  ClausepIChunk*       tl() const;                // last chunk;
  int              one_chunk() const;         // true if hd == tl()

public:

// constructors, etc.

                    ClausepPlex();                  // no-op

  virtual           ~ClausepPlex();

  
// Access functions 
    
  virtual Clausep&      operator [] (int idx) = 0; // access by index;
  virtual Clausep&      operator () (Pix p) = 0;   // access by Pix;

  virtual Clausep&      high_element () = 0;      // access high element
  virtual Clausep&      low_element () = 0;       // access low element

// read-only versions for const Plexes

  virtual const Clausep& operator [] (int idx) const = 0; // access by index;
  virtual const Clausep& operator () (Pix p) const = 0;   // access by Pix;

  virtual const Clausep& high_element () const = 0;      // access high element
  virtual const Clausep& low_element () const = 0;       // access low element


// Index functions

  virtual int       valid (int idx) const = 0;      // idx is an OK index

  virtual int       low() const = 0;         // lowest index or fence if none
  virtual int       high() const = 0;        // highest index or low-1 if none

  int               ecnef() const;         // low limit index (low-1)
  int               fence() const;         // high limit index (high+1)

  virtual void      prev(int& idx) const= 0; // set idx to preceding index
                                          // caution: pred may be out of bounds

  virtual void      next(int& idx) const = 0;       // set to next index
                                          // caution: succ may be out of bounds

  virtual Pix       first() const = 0;        // Pix to low element or 0
  virtual Pix       last() const = 0;         // Pix to high element or 0
  virtual void      prev(Pix& pix) const = 0;  // preceding pix or 0
  virtual void      next(Pix& pix) const = 0;  // next pix or 0
  virtual int       owns(Pix p) const = 0;     // p is an OK Pix

// index<->Pix 

  virtual int       Pix_to_index(Pix p) const = 0;   // get index via Pix
  virtual Pix       index_to_Pix(int idx) const = 0; // Pix via index

// Growth

  virtual int       add_high(const Clausep  elem) =0;// add new element at high end
                                                // return new high

  virtual int       add_low(const Clausep  elem) = 0;   // add new low element,
                                                // return new low

// Shrinkage

  virtual int       del_high() = 0;           // remove the element at high end
                                          // return new high
  virtual int       del_low() = 0;        // delete low element, return new lo

                                          // caution: del_low/high
                                          // does not necessarily 
                                          // immediately call Clausep::~Clausep


// operations on multiple elements

  virtual void      fill(const Clausep  x);          // set all elements = x
  virtual void      fill(const Clausep  x, int from, int to); // fill from to to
  virtual void      clear() = 0;                // reset to zero-sized Plex
  virtual int       reset_low(int newlow); // change low index,return old
  virtual void      reverse();                   // reverse in-place
  virtual void      append(const ClausepPlex& a);    // concatenate a copy
  virtual void      prepend(const ClausepPlex& a);   // prepend a copy

// status

  virtual int       can_add_high() const = 0;
  virtual int       can_add_low() const = 0;
  
  int               length () const;       // number of slots

  int               empty () const;        // is the plex empty?
  virtual int       full() const = 0;      // it it full?

  int               chunk_size() const;    // report chunk size;

  virtual int       OK() const = 0;        // representation invariant

  volatile void     error(const char* msg) const;
  volatile void     index_error() const;
  volatile void     empty_error() const;
  volatile void     full_error() const;
};


// ClausepIChunk ops

inline int ClausepIChunk:: size() const
{
  return top - base;
}


inline int ClausepIChunk:: base_index() const
{
  return base;
}

inline  int ClausepIChunk:: low_index() const
{
  return low;
}

inline  int  ClausepIChunk:: fence_index() const
{
  return fence;
}

inline  int  ClausepIChunk:: top_index() const
{
  return top;
}

inline  Clausep* ClausepIChunk:: pointer_to(int i) const
{
  return &(data[i-base]);
}

inline  int  ClausepIChunk:: index_of(const Clausep* p) const
{
  return ((int)p - (int)data) / sizeof(Clausep) + base;
}

inline  int  ClausepIChunk:: possible_index(int i) const
{
  return i >= base && i < top;
}

inline  int  ClausepIChunk:: possible_pointer(const Clausep* p) const
{
  return p >= data && p < &(data[top-base]);
}

inline  int  ClausepIChunk:: actual_index(int i) const
{
  return i >= low && i < fence;
}

inline  int  ClausepIChunk:: actual_pointer(const Clausep* p) const
{
  return p >= data && p < &(data[fence-base]);
}

inline  int  ClausepIChunk:: can_grow_high () const
{
  return fence < top;
}

inline  int  ClausepIChunk:: can_grow_low () const
{
  return base < low;
}

inline  Clausep* ClausepIChunk:: invalidate()
{
  Clausep* p = data;
  data = 0;
  return p;
}


inline ClausepIChunk* ClausepIChunk::prev() const
{
  return prv;
}

inline ClausepIChunk* ClausepIChunk::next() const
{
  return nxt;
}

inline void ClausepIChunk::link_to_prev(ClausepIChunk* prev)
{
  nxt = prev->nxt;
  prv = prev;
  nxt->prv = this;
  prv->nxt = this;
}

inline void ClausepIChunk::link_to_next(ClausepIChunk* next)
{
  prv = next->prv;
  nxt = next;
  nxt->prv = this;
  prv->nxt = this;
}

inline void ClausepIChunk::unlink()
{
  ClausepIChunk* n = nxt;
  ClausepIChunk* p = prv;
  n->prv = p;
  p->nxt = n;
  prv = nxt = this;
}

inline  int ClausepIChunk:: empty() const
{
  return low == fence;
}

inline  int  ClausepIChunk:: full() const
{
  return top - base == fence - low;
}

inline int ClausepIChunk:: first_index() const
{
  return (low == fence)? fence : low;
}

inline int ClausepIChunk:: last_index() const
{
  return (low == fence)? low - 1 : fence - 1;
}

inline  int  ClausepIChunk:: succ(int i) const
{
  return (i < low) ? low : i + 1;
}

inline  int  ClausepIChunk:: pred(int i) const
{
  return (i > fence) ? (fence - 1) : i - 1;
}

inline  int  ClausepIChunk:: valid_index(int i) const
{
  return i >= low && i < fence;
}

inline  int  ClausepIChunk:: valid_pointer(const Clausep* p) const
{
  return p >= &(data[low - base]) && p < &(data[fence - base]);
}

inline  Clausep* ClausepIChunk:: grow_high ()
{
  if (!can_grow_high()) full_error();
  return &(data[fence++ - base]);
}

inline  Clausep* ClausepIChunk:: grow_low ()
{
  if (!can_grow_low()) full_error();
  return &(data[--low - base]);
}

inline  void ClausepIChunk:: shrink_high ()
{
  if (empty()) empty_error();
  --fence;
}

inline  void ClausepIChunk:: shrink_low ()
{
  if (empty()) empty_error();
  ++low;
}

inline Clausep* ClausepIChunk::first_pointer() const
{
  return (low == fence)? 0 : &(data[low - base]);
}

inline Clausep* ClausepIChunk::last_pointer() const
{
  return (low == fence)? 0 : &(data[fence - base - 1]);
}

inline Clausep* ClausepIChunk::succ(Clausep* p) const
{
  return ((p+1) <  &(data[low - base]) || (p+1) >= &(data[fence - base])) ? 
    0 : (p+1);
}

inline Clausep* ClausepIChunk::pred(Clausep* p) const
{
  return ((p-1) <  &(data[low - base]) || (p-1) >= &(data[fence - base])) ? 
    0 : (p-1);
}


// generic Plex operations

inline ClausepPlex::ClausepPlex() {}

inline int ClausepPlex::chunk_size() const
{
  return csize;
}

inline  int ClausepPlex::ecnef () const
{
  return lo - 1;
}


inline  int ClausepPlex::fence () const
{
  return fnc;
}

inline int ClausepPlex::length () const
{
  return fnc - lo;
}

inline  int ClausepPlex::empty () const
{
  return fnc == lo;
}

inline ClausepIChunk* ClausepPlex::tl() const
{
  return hd->prev();
}

inline int ClausepPlex::one_chunk() const
{
  return hd == hd->prev();
}


#endif
