#include "RexxTokens.h"
#include "RexxClauses.h"
#include "Clause.h"
#include "Express.h"
#include "rexxparse.tok.h"
#include "InternalEnv.h"
#include "SymTabs.h"
#include "SymTab.h"
#include "ArgStr.h"
#include "support.h"
#include "Builtin.h"
#include "MemoryFile.h"
#include <signal.h>
#include "Glob.h"
#include "Errors.h"
#include <GetOpt.h>
#include "Trace.h"

extern Glob glob;

//this is TEMPORARY while implementing TRACE!
int trace_flag = 0;

//extern int pp1_lex(), pp2_lex(); // these are the stages

extern int yyparse(void);
extern int rexx_yylex(void);
extern void rexx_postprocess();

// global clause -- filled in by yyparse()
// could remove this someday
Clause* clause;

// yylex is a wrapper
int yylex(void) {
    while (1) {
	RexxToken& nexttoken = glob.Tokens2->NextToken();
	yylval.symbol.string = nexttoken.Value();
	yylval.symbol.secondary = nexttoken.Secondary();
	if (nexttoken.Type() == TOK_NEWLINE && nexttoken.Secondary() > 0) {
	    glob.LineNumber = nexttoken.Secondary();
	    glob.ClauseNumber = 1;
	}
	return nexttoken.Type();
    }
}

// function that returns the next executable clause
Clause* GetClause() {
    // repeat until we get a valid clause (not a null nor a label)
    while (1) {
	clause = glob.Clauses->NextClause(); // get next clause
	// checking for NULLCLAUSE is probably going to go
	if (!(clause->IsA(TOK_NULLCLAUSE) || clause->IsA(TOK_LABEL))) {
	    // update linenumber
	    glob.LineNumber = clause->GetLineNumber();
	    if (trace_flag) 
		clause->TraceBefore();
	    return clause;
	}
    }
    // shouldn't even get here
    abort();
}

// LoadClauses puts in start where it starts and in end where it ends
//   so we can clean up afterwards
int LoadClauses(int& end) {
    int start = glob.Clauses->Free(); // this is where we're going to start putting clauses
    // we need to call the parser here and load up glob_clauses
    clause = 0;
    while (!yyparse()) {
	if (!clause) {
	    // we've reached the end of file
	    glob.Clauses->AddClause(new EofClause);
	    end = glob.Clauses->Free();
	    return start;
	} else if (!clause->IsA(TOK_NULLCLAUSE)) {
	    clause->SetLineNumber(glob.LineNumber);
	    clause->SetClauseNumber(glob.ClauseNumber);
	    glob.ClauseNumber++;
	    glob.Clauses->AddClause(clause);
	}
        clause = 0;
    }
    // parse error not handled
    abort();
}

int LoadTokens() {
    // glob.Stage1 should be all set up before we call this
    PrepareLexers();

    // does all the tokenizing
    rexx_yylex();
    glob.Tokens->ResetPC();
    
    // insert error checking here for CommentNotTerminated, etc.
    if (glob.pp1_InvalidCharacter) {
	SetError(EInvalidCharacterInProgram);
	return 0;
    }

    if (glob.pp1_StringNotTerminated || glob.pp1_CommentNotTerminated) {
	SetError(EUnmatchedCommentOrQuote);
	return 0;
    }

    if (glob.rexx_InvalidHexOrBinString) {
	SetError(EInvalidHexadecimalOrBinaryString);
	return 0;
    }

    // now we do post-processing of the tokens
    glob.Tokens2->clear();
    rexx_postprocess();
    glob.Tokens2->ResetPC();

    glob.LineNumber = 1;
    glob.ClauseNumber = 1;
    return 1;
}

ArgString command_function() {
    Clause* localclause;
    int ret;

    while (1) {
	localclause = GetClause();
	ret = localclause->Execute();
	if (ret) ret = ProcessPendingSignals();
	// we should trap when ret == 0 and figure out what to do next
	if (!ret) {
	    // need to trap these specially
	    if (IsBreak()) {
		SetError(EInvalidBreak);
	    } else if (IsLeave()) {
		SetError(EInvalidLeaveOrIterate);
	    } else if (IsIterate()) {
		SetError(EInvalidLeaveOrIterate);
	    } else if (IsEof()) {
		SetExit(ArgString());
	    }

	    if (IsExit()) break; // exit while loop
	    else if (IsReturn()) break; // exit while loop
	    else if (IsSignal() && SignalOff(glob.RaisedSignal)) {
		// do traceback here
		if (IsSyntaxError() || IsHaltError()) {
		    // make sure we have the right line number here
		    TraceBack();
		}
		break;
	    } else if (IsSignal() && SignalOn(glob.RaisedSignal) && ActionIsSignal(glob.RaisedSignal)) {
		//cout << "Signal " << GetSignalLabel(glob.RaisedSignal).String() << "\n";
		// de-activate signal
		SetSignalOff(glob.RaisedSignal);
		// we're handling it, so lower it
		LowerSignal(glob.RaisedSignal);
		int BeforeJump = glob.LineNumber;
		// try to reposition
		if (!glob.Clauses->Goto(String(GetSignalLabel(glob.RaisedSignal).String()), GetStart(), GetEnd())) {
		    // oops, another problem while trying to handle this signal
		    SetError(ELabelNotFound);
		    // do a labeled traceback here
		    TraceBack();
		    return ArgString();
		}
		// question is whether we should put this stuff before the jump
		// successfuly jumped to label
		glob.SymbolTables.Write("SIGL", ArgString(BeforeJump));
		if (glob.RaisedSignal == SigSyntax) glob.SymbolTables.Write("RC", ArgString((int)glob.Error));
		 
		SetSignalBeingProcessed(glob.RaisedSignal);
		
	    } else if (IsError()) {
		abort();
		return ArgString();
	    }
	}
	ClearError();
    }
    ArgString retval = ReturnValue();
    return retval;
}

// builtin function
ArgString builtinfunction(String label) {
    glob.BuiltinFunctions.Dispatch(label);
    return ReturnValue();
}

// function -- processes stuff for functions
// we can probably remove this
ArgString function() {
    Clause* localclause;
    int ret;

    while (1) {
	localclause = GetClause();
	ret = localclause->Execute();
	if (ret) ret = ProcessPendingSignals();
	// we should trap when ret == 0 and figure out what to do next
	if (!ret) {
	    // need to trap these specially
	    if (IsBreak()) {
		SetError(EInvalidBreak);
	    } else if (IsLeave()) {
		SetError(EInvalidLeaveOrIterate);
	    } else if (IsIterate()) {
		SetError(EInvalidLeaveOrIterate);
	    } else if (IsEof()) {
		SetExit(ArgString());
	    }

	    if (IsExit()) {
		// need to do something special here
		break;
		// return ArgString();
	    }
	    else if (IsReturn()) break; // exit while loop
	    else if (IsSignal() && SignalOff(glob.RaisedSignal)) break;
	    else if (IsSignal() && SignalOn(glob.RaisedSignal)) {
		// de-activate signal
		SetSignalOff(glob.RaisedSignal);
		LowerSignal(glob.RaisedSignal);
		// try to reposition
		if (!glob.Clauses->Goto(String(GetSignalLabel(glob.RaisedSignal).String()), GetStart(), GetEnd())) {
		    SetError(ELabelNotFound);
		    return ArgString();
		}
	    }
	    else if (IsError()) {
		return ArgString();
	    }
	}
	ClearError();
    }

    ArgString retval = ReturnValue();
    
    return retval;
}

// internal function -- called when we know there is an internal function
ArgString internalfunction(String label) {
    // save old pc
    int OldPC = glob.Clauses->GetPC();
    // position on the new code
    if (!glob.Clauses->Goto(label, GetStart(), GetEnd())) {// do something here
	SetError(ERoutineNotFound);
	return ArgString();
	// couldn't find a label
    }

    ArgString ret = command_function();
    // put back old pc
    glob.Clauses->SetPC(OldPC);
    if (IsExit()) {
	ret = ArgString();
    }
    return ret;
}

ArgString internalfunction(int position) {
    // save old pc
    int OldPC = glob.Clauses->GetPC();
    // position on the new code

    if (!glob.Clauses->Goto(position)) abort();

    ArgString ret = command_function();
    // put back old pc
    glob.Clauses->SetPC(OldPC);
    if (IsExit()) {
	ret = ArgString();
    }
    return ret;
}


ArgString externalfunction(String name, int IsCall) {
    int previousline = glob.LineNumber;

    if (!glob.Stage1.LoadFromFile(name)) {
	fprintf(stderr, "Couldn't load %s\n", (char*)name);
	// should really abort();
	abort();  // this should never happen since we call 'searchforrexxfile()' first!
    }
    if (!LoadTokens()) {
	// we're sure this is a syntax error
	TraceBack();
	ClearError();
	ClearPendingSignals();
	return ArgString();
    }


    int end;
    int start = LoadClauses(end);

    int previous = glob.Clauses->GetPC();
    glob.Clauses->SetPC(start);
    // push an environment
    // arguments have already been set up for us!
    glob.Environments.GetData()->MakeDefault();

    SetStart(start);
    SetEnd(end);

    if (!SourcecodeLoaded()) {
	SetSourcecode(glob.Stage1);
    }

    glob.SymbolTables.PushSymbolTable(); // we need at least one symboltable
    HasSymbolTable(1); // inform env. that we have a symboltable
    // now we need to set up SOURCE

    String nameminuspath = name.after(name.index("/", -1));
    String filename = nameminuspath.before(".");
    String extension = nameminuspath.after(".");
    if (extension == "") extension = "*";
    String source;
    if (IsCall) {
	source = "UNIX SUBROUTINE " + filename + " " + extension + " " + name;
    } else {
	source = "UNIX FUNCTION " + filename + " " + extension + " " + name;
    }
    SetSource(source);

    ArgString ret = command_function(); // process the function

    glob.Clauses->Free(start);
    glob.Clauses->SetPC(previous);
    // our calling routine handles popping the symboltable for us

    // if we exited from this external function, we should convert it into a return
    if (ret.IsValid() && IsExit()) {
	SetReturn(ret);
    }
    
    if (!ret.IsValid()) {
	if (IsExit()) {
	    SetReturn(ArgString());
	    ret = ArgString();
	} else if (IsReturn()) {
	} else {
	    ClearError();
	    SetReturn(ArgString());
	    ret = ArgString();
	}
    }

    ClearPendingSignals();
    glob.LineNumber = previousline;
    return ret;
}

int searchforrexxfile(String name, String extension, String& resolved) {
    FILE* fp;

    if (name.contains("/")) {
	// name is an absolute path
        fp = fopen(name+"."+extension, "r");
	if (fp) {
	    fclose(fp);
	    resolved = name+"."+extension;
	    return 1;
	}
	return 0;
    }

    String env = getenv("PATH");
    String paths[50];
    int n = split(env, paths, 50, String(":"));
    for (int i = 0; i < n; i++) {
	fp = fopen(paths[i]+"/"+name+"."+extension, "r");
	if (fp) {
	    resolved = paths[i]+"/"+name+"."+extension;
	    fclose(fp);
	    return 1;
	}
	fclose(fp);
    }
    // couldn't find it
    return 0;
}

// we still need a 'search for function' function

ArgString function(String label, int IsCall, int skipinternalsearch = 0) {
    String name;
    if (!skipinternalsearch && glob.Clauses->Exists(label, GetStart(), GetEnd())) {
	return internalfunction(label);
    } else if (glob.BuiltinFunctions.Exists(label)) {
	return builtinfunction(label);
    } else if (searchforrexxfile(label, glob.DefaultExtension, name)) {
	return externalfunction(name, IsCall);
    } else {
	SetError(ERoutineNotFound);
	return ArgString(); // send an invalid result string
    }
}

void haltrequested() {
    //    cout << "Halt requested\n";
    SetSignalPending(SigHalt); // this will get caught when we 'processsignalpending'
    return;
}

ArgString externalcommand(String name, ArgString arg) {
    int previousline = glob.LineNumber;

    if (!glob.Stage1.LoadFromFile(name)) {
	fprintf(stderr, "Couldn't load %s\n", (char*)name);
	// should really abort();
	abort();  // this should never happen since we call 'searchforrexxfile()' first!
    }
    if (!LoadTokens()) {
	// for sure, a syntax error
	TraceBack();
	ClearError();
	ClearPendingSignals();
	return ArgString((int)10);
    }

    int end;
    int start = LoadClauses(end);
    int previous = glob.Clauses->GetPC();
    glob.Clauses->SetPC(start);
    // push an environment
    glob.Environments.PushEnv();
    glob.Environments.GetData()->MakeDefault();

    SetStart(start);
    SetEnd(end);

    if (!SourcecodeLoaded()) {
	SetSourcecode(glob.Stage1);
    }

    glob.SymbolTables.PushSymbolTable(); // we need at least one symboltable
    HasSymbolTable(1); // inform env. that we have a symboltable
    // now we need to set up SOURCE
    String nameminuspath = name.after(name.index("/", -1));
    String filename = nameminuspath.before(".");
    String extension = nameminuspath.after(".");
    if (extension == "") extension = "*";
    String source = "UNIX COMMAND " + filename + " " + extension + " " + name;
    SetSource(source);
    if (arg.IsValid()) {
	// we were passed an argument
	Args(0) = arg;
	ArgsOmitted(0) = 0;
    }
    ArgString ret = command_function(); // process the command
    glob.Clauses->Free(start);
    glob.Clauses->SetPC(previous);
    if (HasSymbolTable()) delete glob.SymbolTables.PopSymbolTable(); // pop the one we created for this
    delete glob.Environments.PopEnv();

    if (!ret.IsValid()) {
	// handle the error here, if there is one
	if (IsReturn()) {
	    // don't do anything -- user just returned  with no value
	    ret = ArgString((int)0);
	} else if (IsExit()) {
	    // convert it into a return (user exited with no value)
	    SetReturn(ArgString());
	    ret = ArgString((int)0);
	} else {
	    ClearError();
	    // we've effectively handled this error
	    SetReturn(ArgString());
	    ret = ArgString((int)10);
	}
    }
    ClearPendingSignals(); // make sure there are no signals from this command
    glob.LineNumber = previousline; // restore previous linenumber
    return ret;
}

ArgString externalstring(String string) {
    int previousline = glob.LineNumber;

    glob.Stage1.LoadFromString(string);

    if (!LoadTokens()) {
	// for sure, a syntax error
	TraceBack();
	ClearError();
	ClearPendingSignals();
	return ArgString();
    }

    int end;
    int start = LoadClauses(end);

    int previous = glob.Clauses->GetPC();
    glob.Clauses->SetPC(start);
    // push an environment
    glob.Environments.PushEnv();
    glob.Environments.GetData()->MakeDefault();

    // these tell us where we are legally allowed to search for labels
    SetStart(start);
    SetEnd(end);

    if (!SourcecodeLoaded()) {
	SetSourcecode(glob.Stage1);
    }

    glob.SymbolTables.PushSymbolTable(); // we need at least one symboltable
    HasSymbolTable(1); // inform env. that we have a symboltable
    // now we need to set up SOURCE
    String filename = "*";
    String extension = "*";
    String name = "*";
    String source = "UNIX COMMAND " + filename + " " + extension + " " + name;
    SetSource(source);

    ArgString ret = command_function(); // process the command

    glob.Clauses->Free(start);
    glob.Clauses->SetPC(previous);
    if (HasSymbolTable()) delete glob.SymbolTables.PopSymbolTable(); // pop the one we created for this
    delete glob.Environments.PopEnv();

    if (!ret.IsValid()) {
	// handle the error here, if there is one
	if (IsReturn()) {
	    // don't do anything -- user just returned  with no value
	    ret = ArgString((int)0);
	} else if (IsExit()) {
	    // convert it into a return (user exited with no value)
	    SetReturn(ArgString());
	    ret = ArgString((int)0);
	} else {
	    ClearError();
	    // we've effectively handled this error
	    SetReturn(ArgString());
	    ret = ArgString((int)10);
	}
    }

    ClearPendingSignals();
    glob.LineNumber = previousline;
    return ret;
}

int convertretvalue(ArgString a) {
    if (!a.IsValid()) return 0;
    else if (!a.IsAWhole()) return 10;
    else return a.Integer();
}

void parsename(String& name, String& path, String& filename, String& extension) {
    String filenamewithextension = name.after(name.index("/", -1));
    extension = filenamewithextension.after(".");
    filename = filenamewithextension.before(".");
    if (filename == "") filename = filenamewithextension;
    if (extension == "") extension = "rexx";
    path = name.before(name.index("/", -1));
    if (path != "") path = path + "/";
}

main(int argc, char* argv[]) {
    GetOpt getopt(argc, argv, "tsx:");
    int option_char;
    int iscommand = 1;
    while ((option_char = getopt()) != EOF)
	switch (option_char) {
	  case 't' :
	    trace_flag = 1; // trace stuff
	    break;
	  case 's' :
	    iscommand = 0; // args are a string
	    break;
/*	  case 'x' :
	    // change default extension here
	    glob.DefaultExtension = getopt.optarg;
	    break;
*/
	  case '?' :
	    fprintf(stderr, "usage: %s [-s] arg [args...]\n", argv[0]);
	    exit(10);
	};
    
    if (argc == getopt.optind) {
	// no arguments at all
	fprintf(stderr, "usage: %s [-s] arg [args...]\n", argv[0]);
	exit(10);
    }

    // name should not have an extension
    String name = argv[getopt.optind]; // the name of the invoked program
    String path, filename;
    parsename(name, path, filename, glob.DefaultExtension);

    glob.Init();

    // set up signal handler
    signal(SIGINT, haltrequested);

    int exists;
    ArgString ret;

    if (iscommand) {
	exists = searchforrexxfile(path+filename, glob.DefaultExtension, name);
	if (!exists) {
	    
	    fprintf(stderr, "%s doesn't exist.\n", (char*)name);
	    exit(10);
	}
	if (argc == getopt.optind + 1) {
	    ret = externalcommand(name, ArgString());
	} else {
	    String args;
	    for (int i = getopt.optind + 1; i < argc - 1; i++) args = args + argv[i] + " ";
	    args = args + argv[i];
	    ret = externalcommand(name, ArgString((char*)args));
	}
    } else {
	// arguments are actually the program (STRING FILE)
	String args;
	for (int i = getopt.optind; i < argc - 1; i++) args = args + argv[i] + " ";
	args = args + argv[i];
	ret = externalstring(args);
    }	    


    int intret;

    if (ret.IsValid()) {
	// cout << ret.String() << " <- ReturnValue\n";
	intret = convertretvalue(ret);
    } else {
	// we should NEVER get here
	abort();
	if (IsReturn() || IsExit()) {
	    intret = 0;
	    // we got here because there was no return value supplied
	} else {
	    // print out some nasty errors
	    fprintf(stderr, "+++ Error %d in line %d:%s\n",
		    glob.Error, glob.LineNumber, GetErrorText(glob.Error));
	    intret = 10;
	}
    }

    return intret;
}


